@@ -42,11 +42,15 @@ TOO_BIG_TEXT = '''Maximum workspace size is %.2fMB.\n
 class FlooHandler(base.BaseHandler):
     PROTOCOL = floo_proto.FlooProtocol
 
-    def __init__(self, owner, workspace, auth, upload=None):
-        super(FlooHandler, self).__init__(auth)
+    def __init__(self, owner, workspace, auth, action):
+        self.username = auth.get('username')
+        self.secret = auth.get('secret')
+        self.api_key = auth.get('api_key')
+        # BaseHandler calls reload_settings()
+        super(FlooHandler, self).__init__()
         self.owner = owner
         self.workspace = workspace
-        self.upload_path = upload and utils.unfuck_path(upload)
+        self.action = action
         self.reset()
 
     def _on_highlight(self, data):
@@ -58,6 +62,9 @@ class FlooHandler(base.BaseHandler):
     def get_view(self, buf_id):
         raise NotImplementedError("get_view not implemented")
 
+    def get_view_text_by_path(self, rel_path):
+        raise NotImplementedError("get_view_text_by_path not implemented")
+
     def build_protocol(self, *args):
         self.proto = super(FlooHandler, self).build_protocol(*args)
 
@@ -87,7 +94,7 @@ class FlooHandler(base.BaseHandler):
             'id': buf_id
         })
         buf = self.bufs[buf_id]
-        msg.warn('Syncing buffer %s for consistency.' % buf['path'])
+        msg.warn('Syncing buffer ', buf['path'], ' for consistency.')
         if 'buf' in buf:
             del buf['buf']
 
@@ -136,7 +143,7 @@ class FlooHandler(base.BaseHandler):
         buf_id = data['id']
         buf = self.bufs[buf_id]
         if 'buf' not in buf:
-            msg.debug('buf %s not populated yet. not patching' % buf['path'])
+            msg.debug('buf ', buf['path'], ' not populated yet. not patching')
             return
 
         if buf['encoding'] == 'base64':
@@ -163,14 +170,15 @@ class FlooHandler(base.BaseHandler):
                 buf['buf'] = patch.current
                 buf['md5'] = hashlib.md5(patch.current.encode('utf-8')).hexdigest()
                 buf['forced_patch'] = True
-                msg.debug('forcing patch for %s' % buf['path'])
+                msg.debug('forcing patch for ', buf['path'])
                 self.send(patch.to_json())
                 old_text = view_text
             else:
-                msg.debug('forced patch is true. not sending another patch for buf %s' % buf['path'])
+                msg.debug('forced patch is true. not sending another force patch for buf ', buf['path'])
+
         md5_before = hashlib.md5(old_text.encode('utf-8')).hexdigest()
         if md5_before != data['md5_before']:
-            msg.warn('starting md5s don\'t match for %s. this is dangerous!' % buf['path'])
+            msg.warn('starting md5s don\'t match for ', buf['path'], '. this is dangerous!')
 
         t = DMP.patch_apply(dmp_patches, old_text)
 
@@ -200,22 +208,23 @@ class FlooHandler(base.BaseHandler):
             del buf['timeout_id']
 
         if not clean_patch:
-            msg.log('Couldn\'t patch %s cleanly.' % buf['path'])
+            msg.log('Couldn\'t patch ', buf['path'], ' cleanly.')
             return self.get_buf(buf_id, view)
 
         cur_hash = hashlib.md5(t[0].encode('utf-8')).hexdigest()
         if cur_hash != data['md5_after']:
+            msg.debug('Ending md5s don\'t match for ', buf['path'], ' Setting get_buf timeout.')
             buf['timeout_id'] = utils.set_timeout(self.get_buf, 2000, buf_id, view)
 
         buf['buf'] = t[0]
         buf['md5'] = cur_hash
 
         if not view:
-            msg.debug('No view. Not saving buffer %s' % buf_id)
+            msg.debug('No view. Not saving buffer ', buf_id)
 
             def _on_load():
                 v = self.get_view(buf_id)
-                if v:
+                if v and 'buf' in buf:
                     v.update(buf, message=False)
             self.on_load[buf_id]['patch'] = _on_load
             return
@@ -226,7 +235,7 @@ class FlooHandler(base.BaseHandler):
         buf_id = data['id']
         buf = self.bufs.get(buf_id)
         if not buf:
-            return msg.warn('no buf found: %s.  Hopefully you didn\'t need that' % data)
+            return msg.warn('no buf found: ', data, '. Hopefully you didn\'t need that.')
         timeout_id = buf.get('timeout_id')
         if timeout_id:
             utils.cancel_timeout(timeout_id)
@@ -243,7 +252,7 @@ class FlooHandler(base.BaseHandler):
 
         view = self.get_view(buf_id)
         if not view:
-            msg.debug('No view for buf %s. Saving to disk.' % buf_id)
+            msg.debug('No view for buf ', buf_id, '. Saving to disk.')
             return utils.save_buf(data)
 
         view.update(data)
@@ -255,7 +264,11 @@ class FlooHandler(base.BaseHandler):
             data['buf'] = base64.b64decode(data['buf'])
         self.bufs[data['id']] = data
         self.paths_to_ids[data['path']] = data['id']
-        utils.save_buf(data)
+        view = self.get_view(data['id'])
+        if view:
+            self.save_view(view)
+        else:
+            utils.save_buf(data)
 
     def _on_rename_buf(self, data):
         del self.paths_to_ids[data['old_path']]
@@ -280,7 +293,7 @@ class FlooHandler(base.BaseHandler):
                 del self.paths_to_ids[buf['path']]
                 del self.bufs[buf_id]
         except KeyError:
-            msg.debug('KeyError deleting buf id %s' % buf_id)
+            msg.debug('KeyError deleting buf id ', buf_id)
         # TODO: if data['unlink'] == True, add to ignore?
         action = 'removed'
         path = utils.get_full_path(data['path'])
@@ -289,43 +302,46 @@ class FlooHandler(base.BaseHandler):
             try:
                 utils.rm(path)
             except Exception as e:
-                msg.debug('Error deleting %s: %s' % (path, str_e(e)))
+                msg.debug('Error deleting ', path, ': ', str_e(e))
         user_id = data.get('user_id')
         username = self.get_username_by_id(user_id)
-        msg.log('%s %s %s' % (username, action, path))
+        msg.log(username, ' ', action, ' ', path)
+
+    def _upload_file_by_path(self, rel_path):
+        return self._upload(utils.get_full_path(rel_path), self.get_view_text_by_path(rel_path))
 
     @utils.inlined_callbacks
     def _initial_upload(self, ig, missing_bufs, changed_bufs, cb):
         files, size = yield self.prompt_ignore, ig, G.PROJECT_PATH
 
-        for buf in missing_bufs:
-            self.send({'name': 'delete_buf', 'id': buf['id']})
+        missing_buf_ids = set([buf['id'] for buf in missing_bufs])
+        for buf_id in missing_buf_ids:
+            self.send({'name': 'delete_buf', 'id': buf_id})
 
-        # TODO: pace ourselves (send through the uploader...)
-        for buf in changed_bufs:
-            self.send({
-                'name': 'set_buf',
-                'id': buf['id'],
-                'buf': buf['buf'],
-                'md5': buf['md5'],
-                'encoding': buf['encoding'],
-            })
+        def __upload_buf(buf):
+            return self._upload(utils.get_full_path(buf['path']), buf['buf'])
+
+        changed_bufs_len = reduce(lambda a, buf: a + len(buf.get('buf', '')), changed_bufs, 0)
+        self._rate_limited_upload(iter(changed_bufs), changed_bufs_len, upload_func=__upload_buf)
 
         for p, buf_id in self.paths_to_ids.items():
             if p in files:
                 files.discard(p)
                 continue
+            if buf_id in missing_buf_ids:
+                continue
             self.send({
                 'name': 'delete_buf',
                 'id': buf_id,
             })
 
-        def __upload(relpath):
-            buf_id = self.paths_to_ids.get(relpath)
-            text = self.bufs.get(buf_id, {}).get("buf")
+        def __upload(rel_path):
+            buf_id = self.paths_to_ids.get(rel_path)
+            text = self.bufs.get(buf_id, {}).get('buf')
+            # Only upload stuff that's not in self.bufs (new bufs). We already took care of everything else.
             if text is not None:
                 return len(text)
-            return self._upload(utils.get_full_path(relpath), text)
+            return self._upload(utils.get_full_path(rel_path), self.get_view_text_by_path(rel_path))
 
         self._rate_limited_upload(iter(files), size, upload_func=__upload)
         cb()
@@ -337,6 +353,11 @@ class FlooHandler(base.BaseHandler):
         self.workspace_info = data
         G.PERMS = data['perms']
 
+        self.proto.reset_retries()
+
+        if G.OUTBOUND_FILTERING:
+            msg.error('Detected outbound port blocking! See https://floobits.com/help/network for more info.')
+
         read_only = False
         if 'patch' not in data['perms']:
             read_only = True
@@ -389,14 +410,14 @@ class FlooHandler(base.BaseHandler):
                 buf['view'] = view
                 G.VIEW_TO_HASH[view.native_id] = view_md5
                 if view_md5 == buf['md5']:
-                    msg.debug('md5 sum matches view. not getting buffer %s' % buf['path'])
+                    msg.debug('md5 sum matches view. not getting buffer ', buf['path'])
                 else:
                     changed_bufs.append(buf)
                     buf['md5'] = view_md5
                 continue
 
             try:
-                if buf['encoding'] == "utf8":
+                if buf['encoding'] == 'utf8':
                     if io:
                         buf_fd = io.open(buf_path, 'Urt', encoding='utf8')
                         buf_buf = buf_fd.read()
@@ -411,13 +432,13 @@ class FlooHandler(base.BaseHandler):
                 buf_fd.close()
                 buf['buf'] = buf_buf
                 if md5 == buf['md5']:
-                    msg.debug('md5 sum matches. not getting buffer %s' % buf['path'])
+                    msg.debug('md5 sum matches. not getting buffer ', buf['path'])
                 else:
-                    msg.debug('md5 differs. possibly getting buffer later %s' % buf['path'])
+                    msg.debug('md5 differs. possibly getting buffer later ', buf['path'])
                     changed_bufs.append(buf)
                     buf['md5'] = md5
             except Exception as e:
-                msg.debug('Error calculating md5 for %s, %s' % (buf['path'], str_e(e)))
+                msg.debug('Error calculating md5 for ', buf['path'], ', ', str_e(e))
                 missing_bufs.append(buf)
 
         ignored = []
@@ -426,7 +447,7 @@ class FlooHandler(base.BaseHandler):
                 ignored.append(p)
             new_files.discard(p)
 
-        if self.upload_path:
+        if self.action == utils.JOIN_ACTION.UPLOAD:
             yield self._initial_upload, ig, missing_bufs, changed_bufs
             # TODO: maybe use org name here
             who = 'Your friends'
@@ -436,13 +457,20 @@ class FlooHandler(base.BaseHandler):
             _msg = 'You are sharing:\n\n%s\n\n%s can join your workspace at:\n\n%s' % (G.PROJECT_PATH, who, G.AGENT.workspace_url)
             # Workaround for horrible Sublime Text bug
             utils.set_timeout(editor.message_dialog, 0, _msg)
-
         elif changed_bufs or missing_bufs or new_files:
             # TODO: handle readonly here
-            stomp_local = yield self.stomp_prompt, changed_bufs, missing_bufs, list(new_files), ignored
-            if stomp_local not in [0, 1]:
-                self.stop()
+            if self.action == utils.JOIN_ACTION.PROMPT:
+                stomp_local = yield self.stomp_prompt, changed_bufs, missing_bufs, list(new_files), ignored
+                if stomp_local not in [0, 1]:
+                    self.stop()
+                    return
+            elif self.action == utils.JOIN_ACTION.DOWNLOAD:
+                stomp_local = True
+            else:
+                # This should never happen
+                assert False
                 return
+
             if stomp_local:
                 for buf in changed_bufs:
                     self.get_buf(buf['id'], buf.get('view'))
@@ -478,12 +506,12 @@ class FlooHandler(base.BaseHandler):
             G.PERMS = user_info['perms']
 
     def _on_join(self, data):
-        msg.log('%s joined the workspace' % data['username'])
+        msg.log(data['username'], ' joined the workspace')
         user_id = str(data['user_id'])
         self.workspace_info['users'][user_id] = data
 
     def _on_part(self, data):
-        msg.log('%s left the workspace' % data['username'])
+        msg.log(data['username'], ' left the workspace')
         user_id = str(data['user_id'])
         try:
             del self.workspace_info['users'][user_id]
@@ -521,7 +549,7 @@ class FlooHandler(base.BaseHandler):
         user_id = str(data.get('user_id'))
         username = self.get_username_by_id(user_id)
         if not username:
-            msg.debug('Unknown user for id %s. Not handling request_perms event.' % user_id)
+            msg.debug('Unknown user for id ', user_id, '. Not handling request_perms event.')
             return
 
         perm_mapping = {
@@ -548,7 +576,7 @@ class FlooHandler(base.BaseHandler):
         user_id = str(data['user_id'])
         user = self.workspace_info['users'].get(user_id)
         if user is None:
-            msg.log('No user for id %s. Not handling perms event' % user_id)
+            msg.log('No user for id ', user_id, '. Not handling perms event')
             return
         perms = set(user['perms'])
         if action == 'add':
@@ -596,12 +624,31 @@ class FlooHandler(base.BaseHandler):
             files = files.union(set([utils.to_rel_path(x) for x in ig.files]))
         cb([files, size])
 
-    @utils.inlined_callbacks
     def upload(self, path):
-        files, size = yield self.get_files_from_ignore, path
-        if not (files):
+        if not utils.is_shared(path):
+            editor.error_message('Cannot share %s because is not in shared path %s.\n\nPlease move it there and try again.' % (path, G.PROJECT_PATH))
             return
-        self._rate_limited_upload(iter(files), size)
+        ig = ignore.create_ignore_tree(G.PROJECT_PATH)
+        G.IGNORE = ig
+        is_dir = os.path.isdir(path)
+        if ig.is_ignored(path, is_dir, True):
+            editor.error_message('Cannot share %s because it is ignored.\n\nAdd an exclude rule (!%s) to your .flooignore file.' % (path, path))
+            return
+        rel_path = utils.to_rel_path(path)
+        if not is_dir:
+            self._upload_file_by_path(rel_path)
+            return
+
+        for p in rel_path.split('/'):
+            child = ig.children.get(p)
+            if not child:
+                break
+            ig = child
+
+        if ig.path != path:
+            msg.warn(ig.path, ' is not the same as ', path)
+
+        self._rate_limited_upload(ig.list_paths(), ig.total_size, upload_func=self._upload_file_by_path)
 
     def _rate_limited_upload(self, paths_iter, total_bytes, bytes_uploaded=0.0, upload_func=None):
         reactor.tick()
@@ -638,7 +685,7 @@ class FlooHandler(base.BaseHandler):
                     # work around python 3 encoding issue
                     buf = text.encode('utf8')
                 except Exception as e:
-                    msg.debug('Error encoding buf %s: %s' % (path, str_e(e)))
+                    msg.debug('Error encoding buf ', path, ': ', str_e(e))
                     # We're probably in python 2 so it's ok to do this
                     buf = text
             size = len(buf)
@@ -649,7 +696,7 @@ class FlooHandler(base.BaseHandler):
                 if text is None:
                     buf_md5 = hashlib.md5(buf).hexdigest()
                     if existing_buf['md5'] == buf_md5:
-                        msg.log('%s already exists and has the same md5. Skipping.' % path)
+                        msg.log(path, ' already exists and has the same md5. Skipping.')
                         return size
                     existing_buf['md5'] = buf_md5
                 msg.log('Setting buffer ', rel_path)
@@ -678,7 +725,7 @@ class FlooHandler(base.BaseHandler):
                 buf = base64.b64encode(buf).decode('utf-8')
                 encoding = 'base64'
 
-            msg.log('Creating buffer ', rel_path)
+            msg.log('Creating buffer ', rel_path, ' (', len(buf), ' bytes)')
             event = {
                 'name': 'create_buf',
                 'buf': buf,
@@ -687,9 +734,9 @@ class FlooHandler(base.BaseHandler):
             }
             self.send(event)
         except (IOError, OSError):
-            msg.error('Failed to open %s.' % path)
+            msg.error('Failed to open ', path)
         except Exception as e:
-            msg.error('Failed to create buffer %s: %s' % (path, str_e(e)))
+            msg.error('Failed to create buffer ', path, ': ', str_e(e))
         return size
 
     def stop(self):
